--
--	Wholly
--	Written by scott@mithrandir.com
--
--	This was inspired by my need to replace EveryQuest with something that was more accurate.  The pins
--	was totally inspired by QuestHubber which I found when looking for a replacement for EveryQuest.  I
--	initially made a port of QuestHubber to QuestHubberGrail to make use of the Grail database, but now
--	I have integrated that work into this addon.  Many thanks for all the work put into QuestHubber.
--
--	Version History
--		001	Initial version.
--		002	Support for displaysDungeonQuests now works properly.
--			Added the ability for the panel tooltip to display questgivers.
--			Added the ability to click a quest in the panel to create a TomTom waypoint.
--			Map pins are only displayed for the proper dungeon level of the map.
--		003	Added a panel button to switch to the current zone.
--			Changed the Close button into a Sort button that switches between three different modes:
--				1. Alphabetical  2. Quest level, then alphabetical  3. Quest status, then alphabetical
--			Made map pins only have one pin per NPC, indicating the "best" color possible.
--			The entire zone dropdown button on the quest log panel now can be clicked to change zones.
--			Corrected a problem where someone running without LibStub would get a LUA error.
--			Corrected a localization issue.
--
--	Known Issues
--			Implement the tooltip to do prerequisite information when Grail supports it.
--
--	UTF-8 file
--

local Wholly_File_Version = 003
local requiredGrailVersion = 12

--	Set up the bindings to use the localized name Blizzard supplies.  Note that the Bindings.xml file cannot
--	just contain the TOGGLEQUESTLOG because then the entry for Wholly does not show up.  So, we use a version
--	named WHOLLY_TOGGLEQUESTLOG which maps to the same Global string, which works exactly as we want.
BINDING_NAME_WHOLLY_TOGGLEQUESTLOG = BINDING_NAME_TOGGLEQUESTLOG
BINDING_HEADER_WHOLLY = "Wholly"

if nil == Wholly or Wholly.versionNumber < Wholly_File_Version then

	WhollyDatabase = {}

	Wholly = {

		cachedPanelQuests = {},		-- quests and their status for map area currentPanelMapID
		cachedPinQuests = {},		-- quests and their status for map area currentPinMapID
		color = {
			['B'] = "FF996600",	-- brown	[unobtainable]
			['C'] = "FF00FF00",	-- green	[completed]
			['D'] = "FF0099CC",	-- daily	[repeatable]
			['G'] = "FFFFFF00",	-- yellow	[can accept]
			['I'] = "FFFF00FF",	-- purple	[in log]
			['L'] = "FFFFFFFF",	-- white	[too high level]
			['P'] = "FFFF0000",	-- red		[does not meet prerequisites]
			['R'] = "FF0099CC",	-- daily	[true repeatable - used for question mark in pins]
			['U'] = "FF00FFFF",	-- bogus default[unknown]
			['W'] = "FF666666",	-- grey		[low-level can accept]
			},
		configuration = {
			[1] = {[0]="Prepend quest level", [1]='prependsQuestLevel' },
			[2] = {[0]="Append required level", [1]='appendRequiredLevel' },
			[3] = {[0]="Show repeatable quests", [1]='showsDailyQuests' },
			[4] = {[0]="Show quests in log", [1]='showsQuestsInLog' },
			[5] = {[0]="Show completed quests", [1]='showsCompletedQuests' },
			[6] = {[0]="Show unobtainable quests", [1]='showsUnobtainableQuests' },
			[7] = {[0]="Show low-level quests", [1]='showsLowLevelQuests' },
			[8] = {[0]="Show whether repeatable quests previously completed", [1]='showsAnyPreviousRepeatableCompletions' },
			[9] = {[0]="Show high-level quests", [1]='showsHighLevelQuests'},
			[10]= {[0]="Show quests that need prerequisites", [1]='showsQuestsThatFailPrerequsites'},
			[11]= {[0]="Quest log panel updates when zones change", [1]='updatesPanelWhenZoneChanges'},
			[12]= {[0]="Display map pins for quest givers", [1]='displaysMapPins'},
			[13]= {[0]="Display checkbox on world map", [1]='displaysMapFrame'},
			[14]= {[0]="Display dungeon quests in outer map", [1]='displaysDungeonQuests'},
			},
		currentPanelMapID = nil,	-- the map area the panel is showing
		currentPinMapID = nil,		-- the map area the world map is showing
		currentPinMapLevel = 0,
		currentSortingMode = 1,
		debug = true,
		dropdown = nil,
		eventDispatch = {
			['CHAT_MSG_COMBAT_FACTION_CHANGE'] = function(self, frame)
				self:DelayedUpdateQuestCaches(true)
			end,
			['CHAT_MSG_SKILL'] = function(self, frame)
				self:DelayedUpdateQuestCaches(true)
			end,
			['PLAYER_LEVEL_UP'] = function(self, frame, newLevel)
				self:DelayedUpdateQuestCaches(true)
			end,
			['PLAYER_LOGIN'] = function(self, frame)
				frame:RegisterEvent("CHAT_MSG_COMBAT_FACTION_CHANGE")	-- because specific quests are only available with specific faction reputations
				frame:RegisterEvent("CHAT_MSG_SKILL")			-- because specific quests are only available with specific skill values
				frame:RegisterEvent("PLAYER_LEVEL_UP")			-- because specific quests are only available with specific level values
				frame:RegisterEvent("WORLD_MAP_UPDATE")			-- this is for pins
				frame:RegisterEvent("ZONE_CHANGED_NEW_AREA")		-- this is for the panel
			end,
			['VARIABLES_LOADED'] = function(self, frame)
				if nil == WhollyDatabase.defaultsLoaded then
					self:LoadDefaults()
				end
				self.mapFrame.check:SetChecked(WhollyDatabase.displaysMapPins)
				self:DisplayMapFrame(WhollyDatabase.displaysMapFrame)
				frame:RegisterEvent("PLAYER_LOGIN")
				Grail:RegisterObserverQuestAbandon(self.CallbackHandler)
				Grail:RegisterObserverQuestAccept(self.CallbackHandler)
				Grail:RegisterObserverQuestComplete(self.CallbackHandler)
			end,
			['WORLD_MAP_UPDATE'] = function(self, frame)
				self:UpdateQuestCaches(false, true)	-- note that we do not force the update since this event happens way too many times
			end,
			['ZONE_CHANGED_NEW_AREA'] = function(self, frame)
				self:UpdateQuestCaches(false, false, WhollyDatabase.updatesPanelWhenZoneChanges)
			end,
			},
		filteredPanelQuests = {},	-- filtered table from cachedPanelQuests using current panel filters
		filteredPinQuests = {},		-- filtered table from cachedPinQuests using current pin filters
		initialUpdateProcessed = false,
		mapFrame = nil,			-- the world map frame that contains the checkbox to toggle pins
		npcs = {},
		pairedConfigurationButton = nil,-- configuration panel button that does the same thing as the world map button
		regFrames = {},
		regNPCPins = {},		-- in unifying the pins we keep a table of pins associated with NPC id
		regPins = {},
		s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Dungeon",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			},
		timeSinceLastUpdate = 0,
		tooltip = nil,
		updateDelay = 0.5,
		versionNumber = Wholly_File_Version,

		AttemptUpdateQuestCaches = function(self, elapsed, forceUpdate)
			self.timeSinceLastUpdate = self.timeSinceLastUpdate + elapsed
			if self.timeSinceLastUpdate > self.updateDelay then
				self:UpdateQuestCaches(forceUpdate)
				com_mithrandir_whollyFrame:SetScript("OnUpdate", nil)
				self.timeSinceLastUpdate = 0
			end
		end,

		ButtonClick = function(self, button, ...)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			local locations = Grail:MultipleUniqueMapLocationQuest(button.questId, 'A', nil, nil, true)
			if nil ~= locations then
				for i = 1, #locations do
					local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
					if xcoord ~= nil then
						if realAreaId then npcname = format("%s => %s", npcname, Grail.mapAreaMapping[realAreaId]) end
						TomTom:AddMFWaypoint(mapId, dungeonLevel, xcoord/100, ycoord/100, { title = npcname })
					end
				end
			end
		end,

		ButtonEnter = function(self, button, ...)
			GameTooltip:SetOwner(button, "ANCHOR_RIGHT")
			GameTooltip:ClearLines()
			GameTooltip:SetHyperlink(format("quest:%d", button.questId))

			if self.debug then
				GameTooltip:AddLine(" ")
				GameTooltip:AddDoubleLine("Quest ID:"..button.questId, button.reason)
			end

			local locations = Grail:MultipleUniqueMapLocationQuest(button.questId, 'A', nil, nil, true)
			if nil ~= locations then
				GameTooltip:AddLine(" ")
				GameTooltip:AddLine(self.s['QUESTGIVERS'])
				for i = 1, #locations do
					local xcoord, ycoord, npcName, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
					if xcoord ~= nil then
						if realAreaId then npcname = format("%s => %s", npcname, Grail.mapAreaMapping[realAreaId]) end
						GameTooltip:AddDoubleLine(npcName, Grail.mapAreaMapping[mapId] .. " " .. xcoord .. " " .. ycoord)
					end
				end
			end


			local breadcrumbs = Grail:AvailableBreadcrumbs(button.questId)
			if nil ~= breadcrumbs then
				GameTooltip:AddLine(" ")
				GameTooltip:AddLine(self.s['BREADCRUMB'])
				for i = 1, #breadcrumbs do
					GameTooltip:AddLine(self:PrettyQuestString({ breadcrumbs[i], 'G' }))
				end
			end

			if 'P' == button.statusCode then	-- prerequisites
				-- TODO: Get the list of prerequisite quests and list each in the tooltip with the proper color of the quest (meaning all will be red until the first most likely)
--				GameTooltip:AddLine(" ")
--				GameTooltip:AddLine("Prerequisite Quest(s)")
--				GameTooltip:AddDoubleLine("|c" .. self.color['P'] .. "Some quest" .. "|r", "Some zone name")
			end

			GameTooltip:Show()
		end,

		CallbackHandler = function(type, questId)
			if "Accept" == type then
				Wholly:DelayedUpdateQuestCaches(true)
			else
				Wholly:UpdateQuestCaches(true)
			end
		end,

		CheckNPCTooltip = function(tooltip)
			if (not UnitIsPlayer("mouseover") or true) then
				-- check if this npc drops a quest item
				local guid = UnitGUID("mouseover");
				if ( guid ) then
					local targetType = tonumber(guid:sub(5,5), 16) % 8
					local id = tonumber(guid:sub(7, 10), 16);
					if 1 == targetType then id = id + 1000000 end	-- world object
					local q = Wholly.npcs[id]
					if q and Grail:Status(q, true, nil, nil, nil, nil, true) then
						local _, kindsOfNPC = Grail:IsTooltipNPC(id)
						if nil ~= kindsOfNPC then
							for i = 1, #(kindsOfNPC), 1 do
								local tooltipMessage
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_KILL then tooltipMessage = format(Wholly.s['KILL_TO_START_FORMAT'], Grail:QuestName(q))
								elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then tooltipMessage = format(Wholly.s['DROP_TO_START_FORMAT'], Grail:NPCName(kindsOfNPC[i][2]), Grail:QuestName(q))
								end
								local leftStr = format("|TInterface\\MINIMAP\\ObjectIcons:17:9:0:0:256:128:40:56:32:63|t %s", tooltipMessage)
								tooltip:AddLine(leftStr);
							end
						end
						tooltip:Show();
					end
				end
			end
		end,

		---
		--	Returns the internal classification code for the specified quest
		ClassificationOfQuest = function(self, questId)
			local retval = 'U'
			local questCanAccept, questStatus
			if nil ~= questId and tonumber(questId) then
				questId = tonumber(questId)
				questCanAccept, questStatus = Grail:Status(questId)
				local isDaily = (Grail:IsDaily(questId) or Grail:IsWeekly(questId) or Grail:IsYearly(questId) or Grail:IsRepeatable(questId))
				if "Prerequisites" == questStatus then
					retval = 'P'
				elseif "Level" == questStatus then
					local _, _, _, notToExceed = Grail:MeetsRequirementLevel(questId)
					retval = (UnitLevel('player') > notToExceed) and 'B' or 'L'
				elseif questCanAccept and not isDaily then
					retval = Grail:IsLowLevel(questId) and 'W' or 'G'
				elseif "InLog" == questStatus then
					retval = 'I'
				elseif "Completed" == questStatus then
					retval = 'C'
				else
					local asteriskPosition = strfind(questStatus, '*') or (strlen(questStatus) + 1)
					local s = strsub(questStatus, 1, asteriskPosition - 1)
					if "Class" == s or "Race" == s or "Gender" == s or "Faction" == s or "Invalidated" == s or "Profession" == s or "Reputation" == s or "Holiday" == s then
						retval = 'B'
					elseif isDaily then
						retval = Grail:MeetsRequirementHoliday(questId) and 'D' or 'B'
					else
						retval = 'U'
					end
				end
			end
			if self.debug then
				return retval, questStatus
			else
				return retval
			end
		end,

		---
		--	Gets all the quests in the map area, then classifies them based on the current player.
		ClassifyQuestsInMap = function(self, mapId)
			local retval = nil
			if nil ~= mapId and tonumber(mapId) then
				mapId = tonumber(mapId)
				retval = {}
				local questsInMap = Grail:QuestsInMap(mapId, false, false, WhollyDatabase.displaysDungeonQuests) or {}
				for i = 1, #questsInMap do
					tinsert(retval, { questsInMap[i], self:ClassificationOfQuest(questsInMap[i]) })
				end
			end
			return retval
		end,

		ConfigFrame_OnLoad = function(self, panel)
			panel.name = "Wholly"
			panel:Hide()
			InterfaceOptions_AddCategory(panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				local button = CreateFrame("CheckButton", parent.."Button"..i, panel, "InterfaceOptionsCheckButtonTemplate")
				button:SetPoint("TOPLEFT", panel, "TOPLEFT", 8, i * -20)
				_G[button:GetName().."Text"]:SetText(self.configuration[i][0])
				button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:ScrollFrame_Update() Wholly:UpdatePins() end)

				-- Special case for the button that matches the same functionality as the one on the world map
				if 12 == i then
					self.pairedConfigurationButton = button
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly.mapFrame.check:SetChecked(self:GetChecked()) Wholly:UpdatePins() end)
				end

				-- Special case for the button that controls the display of the world map frame
				if 13 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:DisplayMapFrame(self:GetChecked()) end)
				end
			end
		end,

		ConfigFrame_OnShow = function(self, panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				_G[parent.."Button"..i]:SetChecked(WhollyDatabase[self.configuration[i][1]])
			end
		end,

		DelayedUpdateQuestCaches = function(self, forceUpdate)
			self.timeSinceLastUpdate = 0
			com_mithrandir_whollyFrame:SetScript("OnUpdate", function(myself, elapsed) self:AttemptUpdateQuestCaches(elapsed, forceUpdate) end)
		end,

		DisplayMapFrame = function(self, shouldDisplay)
			if shouldDisplay then self.mapFrame:Show() else self.mapFrame:Hide() end
		end,

		Distance = function(self, x1, y1, x2, y2)
			-- gets distance in map units TODO: use yards?
			local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();

			return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
		end,

		Dropdown_Initialize = function(self, level)
			level = level or 1
			local info
			if 1 == level then
				for i = 1, #(Grail.continents) do
					info = UIDropDownMenu_CreateInfo()
					info.hasArrow = true
					info.notCheckable = true
					info.text = Grail.continents[i]["name"]
					info.value = { ["Level1_Key"] = Grail.continents[i]["name"] }
					UIDropDownMenu_AddButton(info, level)
				end
			elseif 2 == level then
				local Level1_Key = UIDROPDOWNMENU_MENU_VALUE["Level1_Key"]
				local zones = {}
				for i = 1, #(Grail.continents) do
					if Level1_Key == Grail.continents[i]["name"] then
						zones = Grail.continents[i]["zones"]
					end
				end
				for i = 1, #zones do
					info = UIDropDownMenu_CreateInfo()
					info.hasArrow = false
					info.notCheckable = true
					info.text = zones[i]["name"]
					info.value = { ["mapID"] = zones[i]["mapID"] }
					info.func = function()
						Wholly.currentPanelMapID = zones[i]["mapID"]
						Wholly.ForcePanelMapArea(Wholly)
						CloseDropDownMenus()
					end
					UIDropDownMenu_AddButton(info, level)
				end
			end
		end,

		FilterQuests = function(self, forPanel)
			local f = forPanel and self.filteredPanelQuests or self.filteredPinQuests
			f = {}
			local questsInMap = forPanel and self.cachedPanelQuests or self.cachedPinQuests
			local shouldAdd, statusCode
			for i = 1, #questsInMap do
				statusCode = questsInMap[i][2]

				if 'W' == statusCode then shouldAdd = WhollyDatabase.showsLowLevelQuests
				elseif 'I' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsQuestsInLog
				elseif 'C' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsCompletedQuests
				elseif 'B' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsUnobtainableQuests
				elseif 'D' == statusCode then shouldAdd = WhollyDatabase.showsDailyQuests
				elseif 'L' == statusCode then shouldAdd = WhollyDatabase.showsHighLevelQuests
				elseif 'P' == statusCode then shouldAdd = WhollyDatabase.showsQuestsThatFailPrerequsites
				else shouldAdd = true
				end

				if shouldAdd then
					tinsert(f, questsInMap[i])
				end
			end
			if forPanel then self.filteredPanelQuests = f else self.filteredPinQuests = f end
		end,

		FilterPanelQuests = function(self)
			self:FilterQuests(true)
		end,

		FilterPinQuests = function(self)
			self:FilterQuests(false)
		end,

		ForcePanelMapArea = function(self)
			local mapAreaName = Grail.mapAreaMapping[self.currentPanelMapID] or GetRealZoneText()	-- default to something if we do not support the zone
			if nil ~= mapAreaName then UIDropDownMenu_SetText(self.dropdown, mapAreaName) end
			self.cachedPanelQuests = self:ClassifyQuestsInMap(self.currentPanelMapID) or {}
			self:ScrollFrame_Update()
		end,

		GetMousePosition = function(self)
			local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
			local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
			local scale = WorldMapDetailFrame:GetEffectiveScale();

			local x, y = GetCursorPosition();
			local cx = (x/scale - left) / width;
			local cy = (top - y/scale) / height;
	
			return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
		end,

		GetPin = function(self, npcId)
			if nil ~= self.regNPCPins[npcId] then return self.regNPCPins[npcId] end		-- already have a pin for this npcId

			if #self.regFrames > 0 then
				local pin = table.remove(self.regFrames)
				pin.npcId = npcId
			end

			local pin = CreateFrame("Frame", nil, WorldMapDetailFrame);
			pin.npcId = npcId
			pin:SetWidth(16);
			pin:SetHeight(16);
			pin:EnableMouse(true);
			pin:SetScript("OnEnter", function(pin) self:ShowTooltip(pin) end)
			pin:SetScript("OnLeave", function() self:HideTooltip() end)
			pin.SetType = function(self, texType)
				if self.texType == texType then return end -- don't need to make changes

				-- 256x64, 8x4
				self.texture:SetDesaturated(false);
				self.texture:SetVertexColor(1, 1, 1);
				if texType == "D" then
					self.texture:SetTexCoord(3*0.125, 4*0.125, 1*0.250, 2*0.250);
				elseif texType == "R" then
					self.texture:SetTexCoord(4*0.125, 5*0.125, 1*0.250, 2*0.250);
				elseif texType == "P" then
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetVertexColor(1.0, 0.0, 0.0);
				elseif texType == "W" then
					-- I don't like this one, it's too hard to see.
					-- self.texture:SetTexCoord(4*0.125, 5*0.125, 3*0.250, 4*0.250); -- this is the actual dimmed quest giver icon
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetVertexColor(0.75, 0.75, 0.75);
				elseif texType == "L" then
					-- Note: not all systems can desaturate. If they don't, I don't really care.
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetDesaturated(1);
				else
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
				end

				self.texType = texType
			end

			pin.texture = pin:CreateTexture();
			pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
			pin.texture:SetAllPoints();
			pin:SetType("L")
			self.regNPCPins[npcId] = pin
			return pin;
		end,

		HideAllPins = function(self)
			for i, v in pairs(self.regPins) do
				self:HidePin(i)
			end
		end,

		HidePin = function(self, id)
			if self.regPins[id] then
				self.regPins[id]:Hide()
				table.insert(self.regFrames, self.regPins[id])
				self.regNPCPins[self.regPins[id].npcId] = nil
				self.regPins[id] = nil
			end
		end,

		HideTooltip = function(self)
			self.tooltip:Hide()
		end,

		LoadDefaults = function(self)
			local db = {}
			db.defaultsLoaded = true
			db.prependsQuestLevel = true
			db.appendRequiredLevel = true
			db.showsLowLevelQuests = true
			db.showsAnyPreviousRepeatableCompletions = true
			db.updatesPanelWhenZoneChanges = true
			db.displaysMapPins = true
			db.displaysMapFrame = true
			db.displaysDungeonQuests = true
			WhollyDatabase = db
		end,

		OnEvent = function(self, frame, event, ...)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, frame, ...)
			end
		end,

		OnLoad = function(self, frame)
			if not Grail or Grail.versionNumber < requiredGrailVersion then print(format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)) return end
			SlashCmdList["WHOLLY"] = function(msg)
				self:SlashCommand(frame, msg)
			end
			SLASH_WHOLLY1 = "/wholly"
			com_mithrandir_whollyFrameTitleText:SetText("Wholly ".. com_mithrandir_whollyFrameTitleText:GetText())

			com_mithrandir_whollyFrameSortButton:SetText(self.s['SORT'])

			local LibStub = _G["LibStub"]
			if LibStub and 0 < #LibStub then
				local LDB = LibStub("LibDataBroker-1.1")
				if LDB then
					local launcher = LDB:NewDataObject("Wholly", { type="launcher", icon="Interface\\Icons\\INV_Misc_Book_07",
							OnClick = function(theFrame, button) if button == "RightButton" then InterfaceOptionsFrame_OpenToCategory("Wholly") else frame:Show() end end, })
				end
			end

			self.dropdown = CreateFrame("Button", frame:GetName().."ZoneButton", frame, "UIDropDownMenuTemplate")
			UIDropDownMenu_Initialize(self.dropdown, self.Dropdown_Initialize) -- took away "MENU" because no show with it
			self.dropdown:SetPoint("TOPLEFT", frame, "TOPLEFT", 60, -40)
			UIDropDownMenu_SetWidth(self.dropdown, 240, 0)
			UIDropDownMenu_JustifyText(self.dropdown, "LEFT")
			-- By default, the dropdown has it clicking work with the little button on the right.  This makes it work for the whole button:
			self.dropdown:SetScript("OnClick", function(self) ToggleDropDownMenu(nil, nil, Wholly.dropdown) PlaySound("igMainMenuOptionCheckBoxOn") end)

			self.tooltip = CreateFrame("GameTooltip", "com_mithrandir_WhollyTooltip", UIParent, "GameTooltipTemplate");
			self.tooltip:SetFrameStrata("TOOLTIP");
			self.tooltip.large = com_mithrandir_WhollyTooltipTextLeft1:GetFontObject();
			self.tooltip.small = com_mithrandir_WhollyTooltipTextLeft2:GetFontObject();
			self.tooltip.SetLastFont = function(self, fontObj, rightText)
				local txt = rightText and "Right" or "Left"
				_G[format("com_mithrandir_WhollyTooltipText%s%d", txt, self:NumLines())]:SetFont(fontObj:GetFont())
			end

			-- Create the checkbox in the world map
			local frame = CreateFrame("Frame", nil, WorldMapDetailFrame)
			frame:SetSize(250, 20)
			frame:SetPoint("TOPLEFT", WorldMapDetailFrame, "TOPLEFT", 5, -2)
			frame:Hide()
			frame:EnableMouse(1)
			frame:SetFrameStrata("HIGH")
			frame:SetFrameLevel(5)
			frame.check = CreateFrame("CheckButton", nil, frame, "OptionsBaseCheckButtonTemplate")
			frame.check:SetSize(18, 18)
			frame.check:SetHitRectInsets(0, 0, 0, 0)
			frame.check:SetPoint("LEFT")
			frame.check:SetFrameLevel(6)
			frame.check:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[12][1]] = self:GetChecked() Wholly.pairedConfigurationButton:SetChecked(self:GetChecked()) Wholly:UpdatePins() end)
			local font, size = GameTooltipHeader:GetFont()
			frame.text = frame:CreateFontString()
			frame.text:SetFont(font, size, "OUTLINE")
			frame.text:SetTextColor(1, 210/255, 0, 1)
			frame.text:SetText("Wholly")
			frame.text:SetPoint("LEFT", frame.check, "RIGHT")
			self.mapFrame = frame

			-- if the UI panel disappears (maximized WorldMapFrame) we need to change parents
			UIParent:HookScript("OnHide", function()
				self.tooltip:SetParent(WorldMapFrame);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)
			UIParent:HookScript("OnShow", function()
				self.tooltip:SetParent(UIParent);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)

			GameTooltip:HookScript("OnTooltipSetUnit", Wholly.CheckNPCTooltip)

			frame:SetScript("OnEvent", function(frame, event, ...) self:OnEvent(frame, event, ...) end)
			frame:RegisterEvent("VARIABLES_LOADED")

		end,

		---
		--	The first time the panel is shown it is populated with the information from the current map area.
		OnShow = function(self, frame)
			if not self.initialUpdateProcessed then
				self.currentPanelMapID = GetCurrentMapAreaID()
				self:ForcePanelMapArea()
				self.initialUpdateProcessed = true
			end
		end,

		PrettyQuestString = function(self, questTable)
			local questId = questTable[1]
			local filterCode = questTable[2]
			local questLevel = Grail:QuestLevel(questId)
			local questLevelString = WhollyDatabase.prependsQuestLevel and format("[%s] ", questLevel or "??") or ""
			local requiredLevelString = ""
			if WhollyDatabase.appendRequiredLevel then
				local success, _, questLevelNeeded, _ = Grail:MeetsRequirementLevel(questId)
				if not success then requiredLevelString = format(" [%s]", questLevelNeeded) end
			end
			local colorCode = self.color[filterCode]
			local repeatableCompletedString = ('D' == filterCode or 'R' == filterCode) and WhollyDatabase.showsAnyPreviousRepeatableCompletions and Grail:IsResettableQuestCompleted(questId) and "*" or ""
			return format("|c%s%s%s%s%s|r", colorCode, questLevelString, Grail:QuestName(questId), repeatableCompletedString, requiredLevelString)
		end,

		ScrollFrame_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrame_Update
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonTemplate")
		end,

		ScrollFrame_Update = function(self)
			self = self or Wholly
			self:FilterPanelQuests()
			local questsInMap = self.filteredPanelQuests
			local numEntries = #questsInMap
			local buttons = com_mithrandir_whollyFrameScrollFrame.buttons
			local numButtons = #buttons
			local scrollOffset = HybridScrollFrame_GetOffset(com_mithrandir_whollyFrameScrollFrame)
			local buttonHeight = buttons[1]:GetHeight();
			local displayedHeight = 0
			local button, questIndex, questId, questLevelString, requiredLevelString, colorCode, questLevel, filterCode, repeatableCompletedString

			table.sort(questsInMap, Wholly.SortingFunction)
			for i = 1, numButtons do
				button = buttons[i]
				questIndex = i + scrollOffset
				if questIndex <= numEntries then
					questId = questsInMap[questIndex][1]
					filterCode = questsInMap[questIndex][2]
					button.normalText:SetText(self:PrettyQuestString(questsInMap[questIndex]))
					button.questId = questId
					button.statusCode = filterCode
					if self.debug then button.reason = questsInMap[questIndex][3] end
					button:Show()
				else
					button:Hide()
				end
				displayedHeight = displayedHeight + buttonHeight
			end
			HybridScrollFrame_Update(com_mithrandir_whollyFrameScrollFrame, numEntries * buttonHeight, displayedHeight)
		end,

		SetCurrentZone = function(self)
			self:UpdateQuestCaches(false, false, true)
		end,

		ShowPin = function(self, questTable)
			local codeMapping = { ["G"] = 1, ["W"] = 2, ["D"] = 3, ["R"] = 4, ["P"] = 5, ["L"] = 6 }
			local id = questTable[1]
			local locations = Grail:MultipleUniqueMapLocationQuest(id, 'A', self.currentPinMapID, self.currentPinMapLevel)
			if nil ~= locations then
				for i = 1, #locations do
					local xcoord, ycoord, npcName, npcId = locations[i][1], locations[i][2], locations[i][3], locations[i][4]
					if nil ~= xcoord then
						-- record the NPC as needing a tooltip note for the specific quest (it can be a redirect because an actual "NPC" may be the item that starts the quest)
						local shouldProcess, kindsOfNPC = Grail:IsTooltipNPC(npcId)
						if shouldProcess then
							for i = 1, #(kindsOfNPC), 1 do
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_KILL then self.npcs[npcId] = id
								elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then self.npcs[npcId] = id
								elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_BY then self.npcs[tonumber(kindsOfNPC[i][2])] = id
								end
							end
						end

						local pin = self:GetPin(npcId)
						local code = questTable[2]
						if 'D' == code and Grail:IsRepeatable(id) then code = 'R' end
						local codeValue = codeMapping[code]
						local pinValue = codeMapping[pin.texType]
						if codeValue < pinValue then
							pin:SetType(code)
						end						
						pin:ClearAllPoints()
						pin.questId = id
						pin:SetFrameLevel(WorldMapPOIFrame:GetFrameLevel() - 1)
						pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/100*WorldMapDetailFrame:GetWidth(), -ycoord/100*WorldMapDetailFrame:GetHeight())
						pin:Show()

						self.regPins[id..":"..npcId] = pin
					end
				end
			end
		end,

		ShowTooltip = function(self, pin)
			local listedQuests = {}
			self.tooltip:SetOwner(pin, "ANCHOR_RIGHT")
			self.tooltip:ClearLines()
	
			-- find all quests in range of hover
			local mx, my = self:GetMousePosition()
			local npcList = {}
			local npcNames = {}

			local mapId = self.currentPinMapID
			local mapLevel = self.currentPinMapLevel
			local questsInMap = self.filteredPinQuests
			local questId
			for i = 1, #questsInMap do
				questId = questsInMap[i][1]
				local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questId, 'A', mapId, mapLevel)
				if nil ~= locations then
					for j = 1, #locations do
						local xcoord, ycoord, npcname, npc, areaId, dungeonLevel, realAreaId = locations[j][1], locations[j][2], locations[j][3], locations[j][4], locations[j][5], locations[j][6], locations[j][7]
						if nil ~= xcoord then
							local dist = self:Distance(mx, my, xcoord/100, ycoord/100)
							if (dist <= 0.02) then
								if (not npcList[npc]) then
									npcList[npc] = {}
									if mustKill then npcname = format(self.s['MUST_KILL_PIN_FORMAT'], npcname) end
									if realAreaId then npcname = format("%s => %s", npcname, Grail.mapAreaMapping[realAreaId]) end
									npcNames[npc] = npcname
								end
								table.insert(npcList[npc], questsInMap[i])
							end
						end
					end
				end
			end

			local first = true;
			for npc,questList in pairs(npcList) do
				if (not first) then
					self.tooltip:AddLine(" ")
				else
					first = false
				end
				for _,qt in ipairs(questList) do
					local leftStr = self:PrettyQuestString(qt)
					local q = qt[1]
					local escortString = Grail:IsEscort(q) and self.s['ESCORT'] or ""
					local dungeonString = Grail:IsDungeon(q) and self.s['DUNGEON'] or ""
					local raidString = Grail:IsRaid(q) and self.s['RAID'] or ""
					local pvpString = Grail:IsPVP(q) and self.s['PVP'] or ""
					local groupString = Grail:IsGroup(q) and self.s['GROUP'] or ""
					local heroicString = Grail:IsHeroic(q) and self.s['HEROIC'] or ""
					local rightStr = format("%s%s%s%s%s%s", escortString, dungeonString, raidString, pvpString, groupString, heroicString)
					if strlen(rightStr) > 0 then rightStr = format("|c%s%s|r", self.color[qt[2]], rightStr) end

					-- check if already printed - this is for spam quests like the human starting area that haven't been labeled correctly
					if not questName or not listedQuests[questName] then
						self.tooltip:AddDoubleLine(leftStr, rightStr)
						self.tooltip:SetLastFont(self.tooltip.large)
						self.tooltip:SetLastFont(self.tooltip.small, true)
						if questName then listedQuests[questName] = true end
					end
				end
				self.tooltip:AddLine(npcNames[npc], 1, 1, 1, 1)
				self.tooltip:SetLastFont(self.tooltip.small)
			end
	
			self.tooltip:Show();
		end,

		SlashCommand = function(self, frame, msg)
			self:ToggleUI()
		end,

		Sort = function(self)
			-- This is supposed to cycle through the supported sorting techniques and make the contents of the panel
			-- show the quests based on those techniques.
			-- 1 Quest alphabetical
			-- 2 Quest level (and then alphabetical)
			-- 3 Quest type (and then alphabetical)
			self.currentSortingMode = self.currentSortingMode + 1
			if (self.currentSortingMode > 3) then self.currentSortingMode = 1 end
			self:ScrollFrame_Update()
		end,

		SortingFunction = function(a, b)
			local retval = false
			if 1 == Wholly.currentSortingMode then
				retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
			elseif 2 == Wholly.currentSortingMode then
				local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
				if aLevel == bLevel then
					retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
				else
					retval = aLevel < bLevel
				end
			elseif 3 == Wholly.currentSortingMode then
				local aCode, bCode = a[2], b[2]
				if aCode == bCode then
					retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
				else
					retval = aCode < bCode
				end
			end
			return retval
		end,

		ToggleUI = function(self)
			if com_mithrandir_whollyFrame:IsShown() then
				com_mithrandir_whollyFrame:Hide()
			else
				com_mithrandir_whollyFrame:Show()
			end
		end,

		UpdatePins = function(self)
			self:HideAllPins()
			if WhollyDatabase.displaysMapPins then
				self:FilterPinQuests()
				local questsInMap = self.filteredPinQuests
				for i = 1, #questsInMap do
					self:ShowPin(questsInMap[i])
				end
			end
		end,

		UpdateQuestCaches = function(self, forceUpdate, setPinMap, setPanelMap)
			local mapArea = GetCurrentMapAreaID()
			local dungeonLevel = GetCurrentMapDungeonLevel() or 0
			if mapArea ~= self.currentPinMapID or dungeonLevel ~= self.currentPinMapLevel or forceUpdate then
				if setPinMap then self.currentPinMapID = mapArea self.currentPinMapLevel = dungeonLevel end
				self.cachedPinQuests = self:ClassifyQuestsInMap(self.currentPinMapID) or {}
				self:UpdatePins()
			end
			if mapArea ~= self.currentPanelMapID or forceUpdate then
				if setPanelMap then self.currentPanelMapID = mapArea end
				self:ForcePanelMapArea()
			end
		end,

		}

	local locale = GetLocale()
	local C = Wholly.configuration
	if "deDE" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Dungeon",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "esES" == locale or "esMX" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Mazmorra",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "frFR" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Donjon",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "koKR" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "던전",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "ruRU" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Подземелье",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "zhCN" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "Dungeon",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	elseif "zhTW" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show repeatable quests"
		C[4][0]="Show quests in log"
		C[5][0]="Show completed quests"
		C[6][0]="Show unobtainable quests"
		C[7][0]="Show low-level quests"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show quests that need prerequisites"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['DUNGEON'] = "地城",
			['RAID'] = "Raid",
			['PVP'] = "PVP",
			['GROUP'] = "Group",
			['HEROIC'] = "Heroic",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			}
	end

end
